#include "defs.h"
#include <stdio.h>
#include "printtext.h"
#include "imageio.h"
#include "blur.h"
#include "arglist.h"

void printerrorstring(char* title,char *errormessage) {

   if (errormessage!=NULL) fprintf(stderr,"%s: Error: %s!\n\n",title,errormessage);
   fprintf(stderr,"This program will create an image containing\n   highlights of the detected edges in the specified source image!\n");
   fprintf(stderr,"Usage: %s <source1> [<destination>] [--help] [-h]\n\n",title);
}

int main (int argc, char *argv[]) {
   ImagePtr imsource, imdest;	/*Working images*/
   int answer;			/*return value*/
   struct arglist* carglist;	/*commandline options*/

   carglist = arglist_new(argc, argv);
   arglist_addarg (carglist,"--help",0);
   arglist_addarg (carglist,"-h",0);

   printwelcomestring();

   if ((arglist_arggiven(carglist,"--help")==0) || (arglist_arggiven(carglist,"-h")==0)) {
      printerrorstring(argv[0],NULL);
      printusagestring();
      return -1;
   }

   if ((arglist_parameter(carglist,"VOIDARGS",2)!=NULL) || (arglist_parameter(carglist,"VOIDARGS",0)==NULL)){
      printerrorstring(argv[0],"Error parsing commandline");
      printusagestring();
      return -1;
   }
   /*Check parameter count*/

   imsource=loadim(arglist_parameter(carglist,"VOIDARGS",0));
   if (imsource==NULL) {
      printerrorstring(argv[0],"Error loading source image 1");
      printusagestring();
      return -1;
   }
   /*Load Sourcefile*/

   answer=blur(&imdest,imsource);
   /*Do the main work in blur.h*/

   if (arglist_parameter(carglist,"VOIDARGS",1)!=NULL) {
      if (saveim(imdest,arglist_parameter(carglist,"VOIDARGS",1))!=0) {
         printerrorstring(argv[0],"Error saving destination image");
      }
      /*Save the resulting image*/
   }

   destroyim(imsource);
   destroyim(imdest);
   /*We free our memory*/

   fprintf(stderr,"Total amount of blur data is:\n");
   fprintf(stdout,"%i\n",answer);
   
   arglist_kill(carglist);
   /*i forgot to kill the commandline argument list*/

   return 0;

}
