/* imageio: basic image functions*/
/* LibGD Version, fast and optimized*/

#include "defs.h"
#include "gd.h"
#include <stdio.h>
#include <string.h>

/*Internal functions:-------------------------------------------------------*/

inline int fileformatim(char* filename) {
/*Returns 1 for png, 2 for jpg, 3 for wbmp, 4 for gif, */
/*5 for gd, 6 for gd2, 7 for xpm and 8 for xbm*/
   char* fileformat;	/*CharPointer to filetype*/

   fileformat  = strrchr(filename,'.');
   if (fileformat==NULL) {
      fprintf (stderr,"Error: Couldnt determine file extension of %s\n",filename);
      return 0;
   }
   /*Check if the filename has a "." to determine filetype*/

   if (!(strcasecmp(fileformat,".png"))) {
      return 1;
   } else if (!(strcasecmp(fileformat,".jpg")) || !(strcasecmp(fileformat,".jpeg"))) {
      return 2;
   } else if (!(strcasecmp(fileformat,".wbmp"))) {
      return 3;
   } else if (!(strcasecmp(fileformat,".gif"))) {
      return 4;
   } else if (!(strcasecmp(fileformat,".gd"))) {
      return 5;
   } else if (!(strcasecmp(fileformat,".gd2"))) {
      return 6;
   } else if (!(strcasecmp(fileformat,".xpm"))) {
      return 7;
   } else if (!(strcasecmp(fileformat,".xbm"))) {
      return 8;
   } else {
      fprintf (stderr,"Error: Couldnt recognize format %s in %s\n",fileformat,filename);
      return 0;
   }
   /*Check what type of image it is, and if we could possibly open it*/
}


/*External functions:-------------------------------------------------------*/

typedef void* ImagePtr;

ImagePtr loadim(char* filename) {
/*Loads an Image*/
/*Returns gdImagePtr to image structure if succesfull and NULL if not*/

   gdImagePtr im=NULL;		/*ImagePointer to the image*/
   FILE *infile;		/*FilePointer to the image file*/
   int ff;			/*Integer that codes the fileformat type*/

   ff=fileformatim(filename);
   if (ff==0) {
      fprintf(stderr,"Error: Detection of filetype failed\n");
      return NULL;
   }
   /*Check what type of image it is, and if we could possibly open it*/

   if (!(infile  = fopen( filename,"rb"))) {
      fprintf (stderr,"Error: Couldnt open %s\n",filename);
      return NULL;
   }
   /*We have a filename, we know its type, so open it*/
   
   if (ff==1) {
#ifdef USE_PNG_R
      im= gdImageCreateFromPng(infile);
#else
      fprintf (stderr,"Error: Imagetype PNG not supported in this build!");
#endif
   } else if (ff==2) {
#ifdef USE_JPG_R
      im= gdImageCreateFromJpeg(infile);
#else
      fprintf (stderr,"Error: Imagetype JPG not supported in this build!");
#endif
   } else if (ff==3) {
#ifdef USE_WBMP_R
      im= gdImageCreateFromWBMP(infile);
#else
      fprintf (stderr,"Error: Imagetype WBMP not supported in this build!");
#endif
   } else if (ff==4) {
#ifdef USE_GIF_R
      im= gdImageCreateFromGif(infile);
#else
      fprintf (stderr,"Error: Imagetype GIF not supported in this build!");
#endif
   } else if (ff==5) {
#ifdef USE_DGD_R
      im= gdImageCreateFromGd(infile);
#else
      fprintf (stderr,"Error: Imagetype GD not supported in this build!");
#endif
   } else if (ff==6) {
#ifdef USE_DGD2_R
      im= gdImageCreateFromGd2(infile);
#else
      fprintf (stderr,"Error: Imagetype GD2 not supported in this build!");
#endif
   } else if (ff==7) {
#ifdef USE_XPM_R
      im= gdImageCreateFromXpm(filename);
#else
      fprintf (stderr,"Error: Imagetype XPM not supported in this build!");
#endif
   } else if (ff==8) {
#ifdef USE_XBM_R
      im= gdImageCreateFromXbm(infile);
#else
      fprintf (stderr,"Error: Imagetype XBM not supported in this build!");
#endif
   }
   /*Ok now just let libgd create an image out of it*/

   fclose(infile);
   /*Be tidy and close the opened files*/

   return (ImagePtr) im;
   /*Thats all folks*/
}

/*---------------------------------------------------------------------------*/

int saveim(ImagePtr ima, char* filename) {
/*Saves the Image in a file*/
/*Returns null if sucessfull and -1 if not*/

   FILE *outfile;	/*FilePointer to the image file*/
   int ff;		/*Integer that codes the fileformat type*/
   gdImagePtr im;

   im=(gdImagePtr) ima;

   ff=fileformatim(filename);
   if (ff==0) {
      fprintf(stderr,"Error: Detection of filetype failed\n");
      return -1;
   }
   /*Check what type of image it is, and if we could possibly open it*/

   if (!(outfile = fopen(filename,"wb"))) {
      fprintf (stderr,"Error: Couldnt open %s\n",filename);
      return -1;
   }
   /*We have a filename, we know its type, so open it*/

   if (ff==1) {
#ifdef USE_PNG_W
      gdImagePng(im,outfile);
#else
      fprintf (stderr,"Error: Imagetype PNG not supported in this build!");
#endif
   } else if (ff==2) {
#ifdef USE_JPG_W
      gdImageJpeg(im,outfile,-1);
#else
      fprintf (stderr,"Error: Imagetype JPG not supported in this build!");
#endif
   } else if (ff==3) {
#ifdef USE_WBMP_W
      gdImageWBMP(im,gdImageColorAllocate(im, 0, 0, 0),outfile);
#else
      fprintf (stderr,"Error: Imagetype WBMP not supported in this build!");
#endif
   } else if (ff==4) {
#ifdef USE_GIF_W
      gdImageGif(im,outfile);
#else
      fprintf (stderr,"Error: Imagetype GIF not supported in this build!");
#endif
   } else if (ff==5) {
#ifdef USE_DGD_W
      gdImageGd(im,outfile);
#else
      fprintf (stderr,"Error: Imagetype GD not supported in this build!");
#endif
   } else if (ff==6) {
#ifdef USE_DGD2_W
      gdImageGd2(im,outfile,0, GD2_FMT_COMPRESSED);
#else
      fprintf (stderr,"Error: Imagetype GD2 not supported in this build!");
#endif
   } else {
      fprintf (stderr,"Error: Imagetype XPM and XBM not writeable with GD!");
   }
   /*Make libgd save the image to disk*/
   /* TODO: libgd could fail, but i just dunno the return value so
            for now i assume it always works.
            If it doesnt its no problem for me, since
            our program is possibly done with that image >;>>>*/

   fclose(outfile);
   /*be tidy again*/

   return 0;

}

/*---------------------------------------------------------------------------*/

inline ImagePtr createim(int width, int height) {
/*Creates an Image*/
/*Returns ImagePtr to image structure if succesfull and NULL if not*/

return (ImagePtr) gdImageCreate(width,height);

}

/*---------------------------------------------------------------------------*/

inline int destroyim (ImagePtr ima) {
/*Frees image memory*/

   gdImageDestroy((gdImagePtr) ima);
   return 0;
}

/*---------------------------------------------------------------------------*/

inline int widthim(ImagePtr ima) {
/*Returns the images X width*/

   return gdImageSX((gdImagePtr)ima);

}

/*---------------------------------------------------------------------------*/

inline int heightim(ImagePtr ima) {
/*Returns the images X width*/

   return gdImageSY((gdImagePtr)ima);

}

/*---------------------------------------------------------------------------*/

inline int collorallocateim(ImagePtr ima,int r,int g,int b) {
/*Allocates a color structure*/
/*Returns an ID integer of this*/

   return gdImageColorAllocate((gdImagePtr) ima,r,g,b);

}

/*---------------------------------------------------------------------------*/

inline int pixelbrightcim(int r, int g, int b) {
/*Gives back the corresponding pixelbrightness*/
   return ((r+g+b)/3);
}

/*---------------------------------------------------------------------------*/

inline int pixelbrightim(ImagePtr ima,int x,int y) {
/*Gives back the pixelbrightness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   int t;
   gdImagePtr im;
   im=(gdImagePtr) ima;
   
#ifndef skip_boundary_checks
   if (gdImageBoundsSafe(im,x,y)) {
      t = gdImageGetPixel(im,x,y);
   } else {
      return 0;
   }
#else
   if (im->trueColor) {
      t = im->tpixels[y][x];
   } else {
      t = im->pixels[y][x];
   }
#endif
   return pixelbrightcim(gdImageRed(im,t),gdImageGreen(im,t),gdImageBlue(im,t));
}

/*---------------------------------------------------------------------------*/

inline int pixelredim(ImagePtr ima,int x,int y) {
/*Gives back the pixelredness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   int t;
   gdImagePtr im;
   im=(gdImagePtr) ima;

#ifndef skip_boundary_checks
   if (gdImageBoundsSafe(im,x,y)) {
      t = gdImageGetPixel(im,x,y);
   } else {
      return 0;
   }
#else
   if (im->trueColor) {
      t = im->tpixels[y][x];
   } else {
      t = im->pixels[y][x];
   }
#endif
   return (gdImageRed(im,t));
}

/*---------------------------------------------------------------------------*/
inline int pixelgreenim(ImagePtr ima,int x,int y) {
/*Gives back the pixelgreenness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   int t;
   gdImagePtr im;
   im=(gdImagePtr) ima;

#ifndef skip_boundary_checks
   if (gdImageBoundsSafe(im,x,y)) {
      t = gdImageGetPixel(im,x,y);
   } else {
      return 0;
   }
#else
   if (im->trueColor) {
      t = im->tpixels[y][x];
   } else {
      t = im->pixels[y][x];
   }
#endif
   return (gdImageGreen(im,t));
}

/*---------------------------------------------------------------------------*/
inline int pixelblueim(ImagePtr ima,int x,int y) {
/*Gives back the pixelblueness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   int t;
   gdImagePtr im;
   im=(gdImagePtr) ima;

#ifndef skip_boundary_checks
   if (gdImageBoundsSafe(im,x,y)) {
      t = gdImageGetPixel(im,x,y);
   } else {
      return 0;
   }
#else
   if (im->trueColor) {
      t = im->tpixels[y][x];
   } else {
      t = im->pixels[y][x];
   }
#endif
   return (gdImageBlue(im,t));
}

/*---------------------------------------------------------------------------*/

inline int srnddifim(ImagePtr ima,int x, int y) {
/*Gives back the delta pixelbrightness around x,y in im*/
/*Assumes brightness 0 for pixels out of image*/
/* This function returns the maximum brightness difference "around" some pixel
   Pixels with a high brightness change around them are likely to be on
   an "edge" of something*/

   int t,tt,min,max,actual;
   gdImagePtr im;
   im=(gdImagePtr) ima;
   
#ifdef skip_boundary_checks
   int x2,y2;	/*dirty speed hack tempvars*/
#endif
   min=pixelbrightim(im,x,y);
   max=pixelbrightim(im,x,y);
   /*The initial min and max values are to be set to an existant value, so we use the center pixel for this*/

#ifndef skip_boundary_checks
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x+t,y+tt);
         if (actual<min) min=actual;
         if (actual>max) max=actual;
         tt++;
      }
      t++;
   }
   /*go through all 9 pixels around us and check for the darkest
     and the brightest value*/
#else
   x2=x;
   y2=y;
   if (x2==0) x2++;
   if (y2==0) y2++;
   if (x2==gdImageSX(im)-1) x2--;
   if (y2==gdImageSY(im)-1) y2--;
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x2+t,y2+tt);
         if (actual<min) min=actual;
         if (actual>max) max=actual;
         tt++;
      }
      t++;
   }
   /*same as above, but we silently assume an image >=3x3 and do dirty tricks*/ 
#endif

   actual=5*(max-min);
   if (actual>255) actual=255;
   return actual;
   /*Give back 5 times the maximum brightness difference, this way
     greatly outlining all visible structures*/

}
/*---------------------------------------------------------------------------*/

inline int srndavgim(ImagePtr ima,int x, int y) {
/*Gives back the average pixelbrightness around x,y in im*/
/*Assumes brightness 0 for pixels out of image*/

   int t,tt,sum,actual;
   gdImagePtr im;
   im=(gdImagePtr) ima;
   
#ifdef skip_boundary_checks
   int x2,y2;	/*dirty speed hack tempvars*/
#endif
 
#ifndef skip_boundary_checks
   sum=0;
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x+t,y+tt);
	 sum+=actual;
         tt++;
      }
      t++;
   }
   /*go through all 9 pixels around us and add their
     brightness value to sum*/
#else
   x2=x;
   y2=y;
   if (x2==0) x2++;
   if (y2==0) y2++;
   if (x2==gdImageSX(im)-1) x2--;
   if (y2==gdImageSY(im)-1) y2--;
   sum=0;
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x2+t,y2+tt);
	 sum+=actual;
         tt++;
      }
      t++;
   }
   /*same as above but with dirty speedhack, assume image >=3x3*/ 
#endif

   actual=sum/9;
   if (actual>255) actual=255;
   return actual;
   /*Give back the average*/

}
/*---------------------------------------------------------------------------*/

inline int xsectorsim(ImagePtr ima,int sectorsize) {
   int answer; /*return value*/
   gdImagePtr im;
   im=(gdImagePtr) ima;
   answer=( gdImageSX(im)/sectorsize );
   return answer;
}

inline int ysectorsim(ImagePtr ima, int sectorsize) {
   int answer; /*return value*/
   gdImagePtr im;
   im=(gdImagePtr) ima;
   answer=( gdImageSY(im)/sectorsize );
   return answer;
}
/*Give back the amount of sectors in x and y direction in im*/

/*---------------------------------------------------------------------------*/

inline int sectorbrightim(ImagePtr ima,int xpos,int ypos,int sectorsize) {
/*Gives back the average brightness of a specified sector*/
/*As always pixels out of the image are 0*/

   int t,tt,x,y,actual;
   gdImagePtr im;
   im=(gdImagePtr) ima;

   t=0;
   actual=0;
   x=sectorsize*xpos;
   y=sectorsize*ypos;
   while (t<sectorsize) {
      tt=0;
      while (tt<sectorsize){
#ifdef skip_boundary_checks
         if (gdImageBoundsSafe(im,x+t,y+tt))
         /*in case the pixel function doesnt do boundary checking
	   we have to do it :( */
#endif
         actual+=pixelbrightim(im,x+t,y+tt);
         tt++;
      }
      t++;
   }
   return (actual/(sectorsize*sectorsize));
   /*we had width*height pixels, so the average is the sum through that*/

}

/*---------------------------------------------------------------------------*/

inline int drawpixelim(ImagePtr ima,int x,int y,int color) {
/*Draws a Pixel into im*/
/*0 if OK, -1 if failed*/
   gdImagePtr im;
   im=(gdImagePtr) ima;

#ifndef skip_boundary_checks
   if (gdImageBoundsSafe(im,x,y)) {
      gdImageSetPixel(im,x,y,color);
   } else {
      return -1;
   }
#else
   if (im->trueColor) {
      im->tpixels[y][x] = color;
   } else {
      im->pixels[y][x] = color;
   }
#endif
   return 0;

}

/*---------------------------------------------------------------------------*/

inline int drawsectorim(ImagePtr ima,int xpos, int ypos, int color,int sectorsize) {
/*Draws or Outlines a sector in im, not sure yet <g>*/
/*0 if OK, -1 if i am not too lazy to detect errors here <g>*/

   int t,tt,x,y;
   gdImagePtr im;
   im=(gdImagePtr) ima;
   t=0;
   x=xpos*sectorsize;
   y=ypos*sectorsize;
   while (t<sectorsize) {
      tt=0;
      while (tt<sectorsize) {
#ifdef skip_boundary_checks
         if (gdImageBoundsSafe(im,x+t,y+tt))
         /*in case the pixel function doesnt do boundary checking
	   we have to do it :( */
#endif
          drawpixelim(im,x+t,y+tt,color);
         tt++;
      }
      t++;
   }

   return 0;
   /* TODO: be less lazy */
}

