/* imageio: basic image functions*/
/* ImageMagick version, BETA Code*/

#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include "magick/api.h"
#include <string.h>



/*Internal functions:-------------------------------------------------------*/
int magick_initialized=0;
ExceptionInfo	magick_exception;

void init_if_needed() {

   if (magick_initialized==0) {
      magick_initialized=1;
      InitializeMagick("");
      GetExceptionInfo(&magick_exception);
   }

}


/*External functions:-------------------------------------------------------*/

typedef struct {
   Image *image;
   ImageInfo *info;
   int red[256];
   int green[256];
   int blue[256];
   int lastcolor;
} ImagePtr;

ImagePtr* loadim(char* filename) {
/*Loads an Image*/
/*Returns gdImagePtr to image structure if succesfull and NULL if not*/

   ImagePtr *im;		/*ImagePointer to the image*/
   im=(ImagePtr*)malloc(sizeof(ImagePtr));
   init_if_needed();   
   im->info=CloneImageInfo((ImageInfo *) NULL);
   strcpy(im->info->filename,filename);
   im->image=ReadImage(im->info,&magick_exception);
   if (magick_exception.severity != UndefinedException)
       CatchException(&magick_exception);

   im->lastcolor=0;
   if (im->image== NULL) {
      fprintf (stderr,"Error: Couldnt load %s\n",filename);
      return NULL;
   }
   
   return im;
   /*Thats all folks*/
}

/*---------------------------------------------------------------------------*/

int saveim(ImagePtr *im, char* filename) {
/*Saves the Image in a file*/
/*Returns null if sucessfull and -1 if not*/


   strcpy(im->image->filename,filename);
   /*DisplayImages(im->info,im->image);*/
   WriteImage(im->info,im->image);

   /* TODO: ImageMagick could fail, but i just dunno the return value so
            for now i assume it always works.
            If it doesnt its no problem for me, since
            our program is possibly done with that image >;>>>*/

   return 0;

}

/*---------------------------------------------------------------------------*/

inline ImagePtr* createim(int width, int height) {
/*Creates an Image*/
/*Returns ImagePtr to image structure if succesfull and NULL if not*/

   ImagePtr *im;
   
   im=(ImagePtr*)malloc(sizeof(ImagePtr));
   im->info=CloneImageInfo((ImageInfo *) NULL);
   im->image=AllocateImage(im->info);
   im->image->columns=width;
   im->image->rows=height;
   im->lastcolor=0;
   
   if (im->image==NULL) {
      fprintf(stderr,"Error: Couldnt create new image!");
      return NULL;
   }
 
   return im;

}

/*---------------------------------------------------------------------------*/

inline int destroyim (ImagePtr *im) {
/*Frees image memory*/

   DestroyImage(im->image);
   DestroyImageInfo(im->info);
   return 0;
}

/*---------------------------------------------------------------------------*/

inline int widthim(ImagePtr *im) {
/*Returns the images X width*/

   return im->image->columns;

}

/*---------------------------------------------------------------------------*/

inline int heightim(ImagePtr *im) {
/*Returns the images X width*/

   return im->image->rows;

}

/*---------------------------------------------------------------------------*/

inline int collorallocateim(ImagePtr *im,int r,int g,int b) {
/*Allocates a color structure*/
/*Returns an ID integer of this*/

   if (im->lastcolor<255) {
      im->lastcolor++;
      im->red[im->lastcolor]=r;
      im->green[im->lastcolor]=g;
      im->blue[im->lastcolor]=b;
      return im->lastcolor;
   } else {
      return 0;
   }

}

/*---------------------------------------------------------------------------*/

inline int pixelbrightcim(int r, int g, int b) {
/*Gives back the corresponding pixelbrightness*/
   return ((r+g+b)/3);
}

/*---------------------------------------------------------------------------*/

inline int pixelbrightim(ImagePtr *im,int x,int y) {
/*Gives back the pixelbrightness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   PixelPacket t;

   t=GetOnePixel(im->image,x,y);
   return pixelbrightcim(t.red*256/MaxRGB,
      t.green*256/MaxRGB,t.blue*256/MaxRGB);
}

/*---------------------------------------------------------------------------*/

inline int pixelredim(ImagePtr *im,int x,int y) {
/*Gives back the pixelredness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   PixelPacket t;
   
   t=GetOnePixel(im->image,x,y);
   return (t.red*256/MaxRGB);
}

/*---------------------------------------------------------------------------*/
inline int pixelgreenim(ImagePtr *im,int x,int y) {
/*Gives back the pixelgreenness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   PixelPacket t;
   
   t=GetOnePixel(im->image,x,y);
   return (t.green*256/MaxRGB);
}

/*---------------------------------------------------------------------------*/
inline int pixelblueim(ImagePtr *im,int x,int y) {
/*Gives back the pixelblueness at x,y in im*/
/*Gives back 0 if error (x,y out of range or so)*/

   PixelPacket t;
   
   t=GetOnePixel(im->image,x,y);
   return (t.blue*256/MaxRGB);

}

/*---------------------------------------------------------------------------*/

inline int srnddifim(ImagePtr *im,int x, int y) {
/*Gives back the delta pixelbrightness around x,y in im*/
/*Assumes brightness 0 for pixels out of image*/
/* This function returns the maximum brightness difference "around" some pixel
   Pixels with a high brightness change around them are likely to be on
   an "edge" of something*/

   int t,tt,min,max,actual;
   
#ifdef skip_boundary_checks
   int x2,y2;	/*dirty speed hack tempvars*/
#endif
   min=pixelbrightim(im,x,y);
   max=pixelbrightim(im,x,y);
   /*The initial min and max values are to be set to an existant value, so we use the center pixel for this*/

#ifndef skip_boundary_checks
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x+t,y+tt);
         if (actual<min) min=actual;
         if (actual>max) max=actual;
         tt++;
      }
      t++;
   }
   /*go through all 9 pixels around us and check for the darkest
     and the brightest value*/
#else
   x2=x;
   y2=y;
   if (x2==0) x2++;
   if (y2==0) y2++;
   if (x2==widthim(im)-1) x2--;
   if (y2==heightim(im)-1) y2--;
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x2+t,y2+tt);
         if (actual<min) min=actual;
         if (actual>max) max=actual;
         tt++;
      }
      t++;
   }
   /*same as above, but we silently assume an image >=3x3 and do dirty tricks*/ 
#endif

   actual=5*(max-min);
   if (actual>255) actual=255;
   return actual;
   /*Give back 5 times the maximum brightness difference, this way
     greatly outlining all visible structures*/

}
/*---------------------------------------------------------------------------*/

inline int srndavgim(ImagePtr *im,int x, int y) {
/*Gives back the average pixelbrightness around x,y in im*/
/*Assumes brightness 0 for pixels out of image*/

   int t,tt,sum,actual;
   
#ifdef skip_boundary_checks
   int x2,y2;	/*dirty speed hack tempvars*/
#endif
 
#ifndef skip_boundary_checks
   sum=0;
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x+t,y+tt);
	 sum+=actual;
         tt++;
      }
      t++;
   }
   /*go through all 9 pixels around us and add their
     brightness value to sum*/
#else
   x2=x;
   y2=y;
   if (x2==0) x2++;
   if (y2==0) y2++;
   if (x2==widthim(im)-1) x2--;
   if (y2==heightim(im)-1) y2--;
   sum=0;
   t=-1;
   while (t<2) {
      tt=-1;
      while (tt<2) {
         actual=pixelbrightim(im,x2+t,y2+tt);
	 sum+=actual;
         tt++;
      }
      t++;
   }
   /*same as above but with dirty speedhack, assume image >=3x3*/ 
#endif

   actual=sum/9;
   if (actual>255) actual=255;
   return actual;
   /*Give back the average*/

}
/*---------------------------------------------------------------------------*/

inline int xsectorsim(ImagePtr *im,int sectorsize) {
   int answer; /*return value*/
   answer=( widthim(im)/sectorsize );
   return answer;
}

inline int ysectorsim(ImagePtr *im, int sectorsize) {
   int answer; /*return value*/
   answer=( heightim(im)/sectorsize );
   return answer;
}
/*Give back the amount of sectors in x and y direction in im*/

/*---------------------------------------------------------------------------*/

inline int sectorbrightim(ImagePtr *im,int xpos,int ypos,int sectorsize) {
/*Gives back the average brightness of a specified sector*/
/*As always pixels out of the image are 0*/

   int t,tt,x,y,actual;

   t=0;
   actual=0;
   x=sectorsize*xpos;
   y=sectorsize*ypos;
   while (t<sectorsize) {
      tt=0;
      while (tt<sectorsize){
         actual+=pixelbrightim(im,x+t,y+tt);
         tt++;
      }
      t++;
   }
   return (actual/(sectorsize*sectorsize));
   /*we had width*height pixels, so the average is the sum through that*/

}

/*---------------------------------------------------------------------------*/

inline int drawpixelim(ImagePtr *im,int x,int y,int color) {
/*Draws a Pixel into im*/
/*0 if OK, -1 if failed*/

   PixelPacket *t;

   t=SetImagePixels(im->image,x,y,1,1);
   t->red=im->red[color]*MaxRGB/256;
   t->green=im->green[color]*MaxRGB/256;
   t->blue=im->blue[color]*MaxRGB/256;
   
   return 0;

}

/*---------------------------------------------------------------------------*/

inline int drawsectorim(ImagePtr *im,int xpos, int ypos, int color,int sectorsize) {
/*Draws or Outlines a sector in im, not sure yet <g>*/
/*0 if OK, -1 if i am not too lazy to detect errors here <g>*/

   int t,tt,x,y;
   t=0;
   x=xpos*sectorsize;
   y=ypos*sectorsize;
   while (t<sectorsize) {
      tt=0;
      while (tt<sectorsize) {
          drawpixelim(im,x+t,y+tt,color);
         tt++;
      }
      t++;
   }

   return 0;
   /* TODO: be less lazy */
}

